from unittest.mock import Mock

from pypck.pck_commands import PckParser
from tests.conftest import encode_pck, password, username

# Socket connection tests


def test_called_connection_made(pchk_connection_manager):
    """Test if socket_connected future is done after connection to socket
    is established.
    """
    pchk_connection_manager.connect()
    assert pchk_connection_manager.socket_connected.done()


def test_async_connect(monkeypatch, loop, pchk_connection_manager):
    """Tests if async_connect coroutine continues after all connection
    futures are done.
    """
    monkeypatch.setattr(pchk_connection_manager, 'connect', Mock())
    pchk_connection_manager.socket_connected.set_result(True)
    pchk_connection_manager.lcn_connected.set_result(True)
    pchk_connection_manager.segment_scan_completed.set_result(True)
    loop.run_until_complete(
        pchk_connection_manager.async_connect(timeout=0.5))
    assert pchk_connection_manager.connect.called

# Authentication tests


def test_received_auth_username(pchk_connection_manager):
    """Test username authentication workflow.
    """
    pchk_connection_manager.data_received(
        encode_pck(PckParser.AUTH_USERNAME))
    pchk_connection_manager.send_command.assert_called_with(username)


def test_received_auth_password(pchk_connection_manager):
    """Test password authentication workflow.
    """
    pchk_connection_manager.data_received(
        encode_pck(PckParser.AUTH_PASSWORD))
    pchk_connection_manager.send_command.assert_called_with(password)


def test_received_auth_ok(monkeypatch, pchk_connection_manager):
    """Test authentication ok workflow.
    """
    monkeypatch.setattr(pchk_connection_manager, 'on_auth_ok', Mock())
    pchk_connection_manager.data_received(
        encode_pck(PckParser.AUTH_OK))
    assert pchk_connection_manager.on_auth_ok.called

# LCN Connection tests


def test_received_lcn_connected(monkeypatch, pchk_connection_manager):
    """Test if correct workflow is done after lcn connected message is
    received.
    """
    monkeypatch.setattr(pchk_connection_manager, 'on_successful_login', Mock())
    pchk_connection_manager.data_received(
        encode_pck(PckParser.LCNCONNSTATE_CONNECTED))
    assert pchk_connection_manager.on_successful_login.called

    pchk_connection_manager.send_command.assert_called_with('!OM0P')


def test_called_on_successful_login(monkeypatch, pchk_connection_manager):
    """Test workflow after on_successful_login was called (e.g.
    lcn_connected future is done, ping procedure started, segment scan
    started).
    """
    with monkeypatch.context() as mpc:
        mpc.setattr(pchk_connection_manager, 'ping', Mock())
        mpc.setattr(pchk_connection_manager, 'status_segment_scan', Mock())
        pchk_connection_manager.on_successful_login()

        # assert that lcn_connected future has result set
        assert pchk_connection_manager.lcn_connected.done()

        # assert that TimeoutRetryhandlers for ping and segment_scans are
        # activated
        assert pchk_connection_manager.ping.activate.called
        assert pchk_connection_manager.status_segment_scan.activate.called

# LCN segment scan tests


def test_called_segment_scan(monkeypatch, pchk_connection_manager):
    """Test workflow after segment_scan completed (not) successful.
    """
    monkeypatch.setattr(pchk_connection_manager, 'set_local_seg_id', Mock())
    # assert that for each timeout a segment scan command is sent
    pchk_connection_manager.segment_scan_timeout(False)
    pchk_connection_manager.send_command.assert_called_with('>G003003.SK')

    # assert that if max retries is reached, local segment id is set to 0
    pchk_connection_manager.segment_scan_timeout(True)
    pchk_connection_manager.set_local_seg_id.assert_called_with(0)


def test_received_segment_info(monkeypatch, pchk_connection_manager):
    """Test if local segment id is about to set after receiving
    appropriate PCK command.
    """
    monkeypatch.setattr(pchk_connection_manager, 'set_local_seg_id', Mock())
    pck = '=M000005.SK7'
    pchk_connection_manager.data_received(encode_pck(pck))

    # assert that local segment id is set properly
    pchk_connection_manager.set_local_seg_id.assert_called_with(7)


def test_called_local_seg_id(pchk_connection_manager):
    """Test if local segment id was set correctly after calling
    set_local_seg_id.
    """
    pchk_connection_manager.set_local_seg_id(7)

    assert pchk_connection_manager.local_seg_id == 7
    assert pchk_connection_manager.segment_scan_completed.done()



"""Copyright (c) 2006-2018 by the respective copyright holders.

All rights reserved. This program and the accompanying materials
are made available under the terms of the Eclipse Public License v1.0
which accompanies this distribution, and is available at
http://www.eclipse.org/legal/epl-v10.html

Contributors:
  Andre Lengwenus - port to Python and further improvements
  Tobias Juettner - initial LCN binding for openHAB (Java)
"""

from pypck import (connection, inputs, lcn_addr, lcn_defs, module,
                   pck_commands, timeout_retry)

__all__ = [connection, inputs, lcn_addr, lcn_defs, module, pck_commands,
           timeout_retry]



"""StatusRequestHandler tests."""

import asyncio
from unittest.mock import Mock

import pytest

from pypck import lcn_defs
from pypck.module import ModuleConnection


class MockPchkConnectionManager(Mock):

    def __init__(self):
        super().__init__()
        self.settings = lcn_defs.default_connection_settings
        self.socket_connected = asyncio.Future()
        self.lcn_connected = asyncio.Future()
        self.segment_scan_completed = asyncio.Future()

        self.socket_connected.set_result(True)
        self.lcn_connected.set_result(True)
        self.segment_scan_completed.set_result(True)


@pytest.fixture
def pchk_connection_manager():
    pchk_connection_manager = MockPchkConnectionManager()
    return pchk_connection_manager


@pytest.fixture
def module_connection(monkeypatch, loop, pchk_connection_manager):
    module_connection = ModuleConnection(loop, pchk_connection_manager,
                                         0, 7, sw_age=0x170206)
    monkeypatch.setattr(module_connection, 'send_command', Mock())

    yield module_connection
    loop.run_until_complete(module_connection.cancel_status_request_handlers())


def test_activate_status_outputs(loop, module_connection):
    """Test workflow:
    activate status_request_handler(item) --> send appropriate command
    """
    for item in lcn_defs.OutputPort:
        loop.run_until_complete(
            module_connection.activate_status_request_handler(item))
        module_connection.send_command.assert_called_with(
            False, 'SMA{:d}'.format(item.value + 1))
        module_connection.send_command.reset_mock()
        loop.run_until_complete(
            module_connection.cancel_status_request_handler(item))


def test_activate_status_relays(loop, module_connection):
    """Test workflow:
    activate status_request_handler(item) --> send appropriate command
    """
    for item in lcn_defs.RelayPort:
        loop.run_until_complete(
            module_connection.activate_status_request_handler(item))
        module_connection.send_command.assert_called_with(False, 'SMR')
        module_connection.send_command.reset_mock()
        loop.run_until_complete(
            module_connection.cancel_status_request_handler(item))


def test_activate_status_motor_ports(loop, module_connection):
    """Test workflow:
    activate status_request_handler(item) --> send appropriate command
    """
    for item in lcn_defs.MotorPort:
        loop.run_until_complete(
            module_connection.activate_status_request_handler(item))
        module_connection.send_command.assert_called_with(False, 'SMR')
        module_connection.send_command.reset_mock()
        loop.run_until_complete(
            module_connection.cancel_status_request_handler(item))


def test_activate_status_bin_sensor_ports(loop, module_connection):
    """Test workflow:
    activate status_request_handler(item) --> send appropriate command
    """
    for item in lcn_defs.BinSensorPort:
        loop.run_until_complete(
            module_connection.activate_status_request_handler(item))
        module_connection.send_command.assert_called_with(False, 'SMB')
        module_connection.send_command.reset_mock()
        loop.run_until_complete(
            module_connection.cancel_status_request_handler(item))


def test_activate_status_led_ports(loop, module_connection):
    """Test workflow:
    activate status_request_handler(item) --> send appropriate command
    """
    for item in lcn_defs.LedPort:
        loop.run_until_complete(
            module_connection.activate_status_request_handler(item))
        # print(send_command.call_args_list)
        module_connection.send_command.assert_called_with(False, 'SMT')
        module_connection.send_command.reset_mock()
        loop.run_until_complete(
            module_connection.cancel_status_request_handler(item))


def test_activate_status_keys(loop, module_connection):
    """Test workflow:
    activate status_request_handler(item) --> send appropriate command
    """
    for item in lcn_defs.Key:
        loop.run_until_complete(
            module_connection.activate_status_request_handler(item))
        module_connection.send_command.assert_called_with(False, 'STX')
        module_connection.send_command.reset_mock()
        loop.run_until_complete(
            module_connection.cancel_status_request_handler(item))



"""Copyright (c) 2006-2018 by the respective copyright holders.

All rights reserved. This program and the accompanying materials
are made available under the terms of the Eclipse Public License v1.0
which accompanies this distribution, and is available at
http://www.eclipse.org/legal/epl-v10.html

Contributors:
  Andre Lengwenus - port to Python and further improvements
  Tobias Juettner - initial LCN binding for openHAB (Java)
"""

import logging

import asyncio

_LOGGER = logging.getLogger(__name__)

# The default timeout to use for requests. Worst case: Requesting threshold
# 4-4 takes at least 1.8s
DEFAULT_TIMEOUT_MSEC = 3500


class TimeoutRetryHandler():
    """Manage timeout and retry logic for an LCN request.

    :param           loop:           The asyncio event loop
    :param    int    num_tries:      The maximum number of tries until the
                                     request is marked as failed (-1 means
                                     forever)
    :param    int    timeout_msec:   Default timeout in milliseconds

    """

    def __init__(self, loop, num_tries=3, timeout_msec=DEFAULT_TIMEOUT_MSEC):
        """Construct TimeoutRetryHandler."""
        self.loop = loop
        self.num_tries = num_tries
        self.timeout_msec = timeout_msec
        self._timeout_callback = None
        self.timeout_loop_task = None

    def set_timeout_msec(self, timeout_msec):
        """Set the timeout in milliseconds.

        :param    int    timeout_msec:    The timeout in milliseconds
        """
        self.timeout_msec = timeout_msec

    def set_timeout_callback(self, timeout_callback):
        """Timeout_callback function is called, if timeout expires.

        Function has to take one argument:
        Returns failed state (True if failed)
        """
        self._timeout_callback = timeout_callback

    def activate(self, timeout_callback=None):
        """Schedule the next request."""
        if self.is_active():
            self.loop.create_task(self.cancel())
        if timeout_callback is not None:
            self.set_timeout_callback(timeout_callback)

        self.timeout_loop_task = self.loop.create_task(self.timeout_loop())

    async def done(self):
        """Signal the completion of the TimeoutRetryHandler."""
        await self.timeout_loop_task

    async def cancel(self):
        """Must be called when a response (requested or not) is received."""
        if self.timeout_loop_task is not None:
            self.timeout_loop_task.cancel()
            try:
                await self.timeout_loop_task
            except asyncio.CancelledError:
                pass

    def is_active(self):
        """Check whether the request logic is active."""
        if self.timeout_loop_task is None:
            return False
        return not self.timeout_loop_task.done()

    async def on_timeout(self, failed=False):
        """Is called on timeout of TimeoutRetryHandler."""
        if self._timeout_callback is not None:
            self._timeout_callback(failed)

    async def timeout_loop(self):
        """Timeout / retry loop."""
        tries_left = self.num_tries
        while (tries_left > 0) or (tries_left == -1):
            if not self.timeout_loop_task.done():
                await self.on_timeout()
                await asyncio.sleep(self.timeout_msec / 1000)
                if self.num_tries != -1:
                    tries_left -= 1
            else:
                break

        if not self.timeout_loop_task.done():
            await self.on_timeout(failed=True)
